
{-

timeline az egy ketdimenzios, atfedes nelkuli 
intervallumokat tartalmazo ojjektum

TODO: implement minimum length (when auto-resizing)
-}

{-# LANGUAGE MultiParamTypeClasses, TypeSynonymInstances #-}
module Data.Timeline
  ( Iv
  , Timeline(..)
  , SingleLine(..)
  , Block(..)
  , _block_duration 
  , tlNumberOfRows
  , tlDuration
  , emptyTimeline
  , tlNames
  , IvMap(..)
  , window
  , insertBlock
  , removeBlock
  , moveBlock
  , Alignment(..)
--  , adjustBlock
  )
  where -- Timeline.Pure where

import Control.Monad
import Data.Array.IArray
import Data.Maybe
import Data.Ord
import Data.List hiding (intersect,insert)

-------------------------------------------------------

-- monad-like stuff for Either

liftRight :: (a -> b) -> Either c a -> Either c b
liftRight _ (Left  x) = Left  x
liftRight f (Right y) = Right (f y) 

bindRight :: Either c a -> (a -> Either c b) -> Either c b
bindRight (Left  x) _ = Left x
bindRight (Right y) f = f y

-------------------------------------------------------

data Alignment 
  = NoAlign
  | ResizeLeft   
  | ResizeRight
  | Move
  | FitIntoGap

-------------------------------------------------------

adjustElem :: (IArray a e, Ix i) => (e -> e) -> i -> a i e -> a i e
adjustElem f i ar = ar // [ ( i , f (ar!i) ) ]

adjustElemMaybe :: (IArray a e, Ix i) => (e -> Maybe e) -> i -> a i e -> Maybe (a i e)
adjustElemMaybe f i ar = case f (ar!i) of
  Nothing -> Nothing
  Just xx -> Just $ ar // [ ( i , xx ) ]

adjustElemEither :: (IArray a e, Ix i) => (e -> Either c e) -> i -> a i e -> Either c (a i e)
adjustElemEither f i ar = case f (ar!i) of
  Left err -> Left err
  Right xx -> Right $ ar // [ ( i , xx ) ]

-------------------------------------------------------

type Iv t = (t,t)

data Timeline a t = Timeline
  { _tliv  :: Iv t
  , _tlarr :: Array Int (SingleLine a t) 
  }

newtype SingleLine a t = SL [Block a t]  -- sorted list
 
data Block a t = Block 
  { _block_iv    :: Iv t
  , _block_name  :: String
  , _block_color :: (Float,Float,Float) 
  , _block_user  :: a
  } 
  
_block_duration :: Num t => Block a t -> t
_block_duration block = let (s,t) = _block_iv block in t-s 

tlNumberOfRows :: Timeline a t -> Int
tlNumberOfRows = snd . bounds . _tlarr

tlDuration :: Num t => Timeline a t -> t
tlDuration = ( \(a,b) -> b-a ) . _tliv

-------------------------------------------------------

emptySingleLine :: SingleLine a t
emptySingleLine = SL []

emptyTimeline :: Iv t -> (Int,Int) -> Timeline a t 
emptyTimeline full bounds = 
  Timeline full (listArray bounds $ repeat emptySingleLine)

tlNames :: Timeline a t -> Array Int [String]
tlNames (Timeline iv arr) = amap f arr where
  f (SL xs) = nub $ map _block_name xs
  
-------------------------------------------------------

-- * time maps

{- ezt nem hagyja a GHC
instance Functor Iv where
  fmap f (x,y) = (f x, f y)
-}

onboth :: (a -> b) -> (a,a) -> (b,b)
onboth f (x,y) = (f x, f y)

class IvMap c where
  ivmap :: (Iv s -> Iv t) -> c s -> c t
  
instance IvMap (Block a) where
  ivmap f block@(Block iv nam col usr) = Block (f iv) nam col usr

instance IvMap (SingleLine a) where
  ivmap f (SL xs) = SL $ map (ivmap f) xs

instance IvMap (Timeline a) where
  ivmap f (Timeline iv arr) = Timeline (f iv) (amap (ivmap f) arr) 
  
----------  
  
instance Functor (Block a) where
  fmap f block@(Block iv nam col usr) = Block (onboth f iv) nam col usr

-- [a] is an instance by default
instance Functor (SingleLine a) where
  fmap f (SL xs) = SL $ map (fmap f) xs

instance Functor (Timeline a) where
  fmap f (Timeline iv arr) = Timeline (onboth f iv) (amap (fmap f) arr) 
  
-------------------------------------------------------

-- * windowing 

collapseEither :: Either t t -> t
collapseEither (Left  x) = x
collapseEither (Right y) = y

eitherIntersect :: Ord t => Iv t -> Iv t -> Maybe (Iv (Either t t))
eitherIntersect (s,t) (x,y) 
  | y<=s       = Nothing  
  | x>=t       = Nothing
  | otherwise  = Just (lr u x , lr v y)
  where
    u = max x s
    v = min y t
    lr e f = if e == f then Right e else Left e

-- Left  -> elvagtuk (nem az eredeti boundary)
-- Right -> nem vagtuk el (eredeti boundary)
windowSingleLine' :: Ord t => Iv t -> SingleLine a t -> SingleLine a (Either t t) 
windowSingleLine' win (SL xs) = SL (worker win xs) where
  worker _ [] = []
  worker win@(s,t) (b:bs)  
    | y<=s      = worker win bs  
    | x>=t      = []  
    | otherwise = b' : worker win bs 
    where
      xy@(x,y) = _block_iv b
      b' = ivmap (const is) b
      is = fromJust $ eitherIntersect win xy

window' :: Ord t => Iv t -> Timeline a t -> Timeline a (Either t t) 
window' win (Timeline iv arr) = Timeline iv' arr' where
  iv'  = fromJust $ eitherIntersect win iv
  arr' = amap (windowSingleLine' win) arr
   
window :: Ord t => Iv t -> Timeline a t -> Timeline a t 
window win tline = fmap collapseEither (window' win tline)

-------------------------------------------------------

-- * gaps

-- calculates all the gaps between nonoverlapping closed /sorted/ intervals
-- @gaps biginterval [smallintervsals]@
calcGaps :: Ord t => Iv t -> [Iv t] -> [Iv t]
calcGaps (a,b) [] = if b>a then [(a,b)] else []
calcGaps (a,b) ((x,y):rest)  
  = ( if x>a then ((a,x):) else id )
  $ calcGaps (y,b) rest
   
firstGap :: Ord t => Iv t -> [Iv t] -> Maybe (Iv t)
firstGap ab xys = case calcGaps ab xys of
  [] -> Nothing
  (xy:_) -> Just xy
  
isSubInterval :: Ord t => Iv t -> Iv t -> Bool
isSubInterval small@(x,y) big@(a,b) = (x>=a) && (y<=b)
  
doesIntersect :: Ord t => Iv t -> Iv t -> Bool
doesIntersect (a,b) (c,d) = not ( (b<=c) || (d<=a) )
  
-- second argument is a list of GAPS
doesFitIntoAGap :: Ord t => Iv t -> [Iv t] -> Bool  
doesFitIntoAGap small gaps = or $ map (isSubInterval small) gaps

-- second argument is a list of GAPS
closestGap :: Ord t => Iv t -> [Iv t] -> Maybe (Iv t)
closestGap iv gaps = case gaps of
  [] -> Nothing 
  _  -> case find (doesIntersect iv) gaps of
    Just gap -> Just gap
    Nothing  -> Nothing -- Just (head gaps)
  
-------------------------------------------------------
  
removeSingleLine :: Ord t => Iv t -> SingleLine a t -> ( Maybe (Block a t) , SingleLine a t ) 
removeSingleLine st@(s,t) (SL xs) = ( mblock , SL line ) where
  ( mblock , line ) = worker Nothing xs
  worker m [] = ( m , [] )
  worker m bbs@(b:bs) 
    | t <= x     = ( Nothing , bbs )
    | xy == st   = ( Just b  , bs  )
    | otherwise  = let (m',ys) = worker m bs in (m',b:ys)
    where
      xy@(x,y) = _block_iv b

removeBlock :: Ord t => Int -> Iv t -> Timeline a t -> ( Maybe (Block a t) , Timeline a t )
removeBlock lev st (Timeline iv arr) = ( mblock , Timeline iv arr' ) where
  (mblock , line') = removeSingleLine st (arr!lev) 
  arr' = case mblock of 
    Nothing -> arr 
    Just _ -> adjustElem (const line') lev arr

{-
insertSingleLine :: Ord t => Iv t -> Bool -> Block a t -> SingleLine a t -> Either String (SingleLine a t)
insertSingleLine full fitinto block (SL xs) = liftRight SL result where
  st = _block_iv block
  gaps = calcGaps full (map _block_iv xs) 
  b = doesFitIntoAGap st gaps
  result = case (fitinto,b) of
    (_    ,True ) -> Right $ worker st block xs
    (False,False) -> Left "does not fit into a gap"
    (True ,False) -> case closestGap st gaps of
      Nothing  -> Left "there is no gap intersecting"
      Just new -> Right $ worker new (ivmap (const new) block) xs
  worker uv@(u,v) block [] = [block]
  worker uv@(u,v) block bbs@(b:bs) 
    | v<=x       = block : bbs
    | otherwise  = b : worker uv block bs
    where
      xy@(x,y) = _block_iv b
-}

insertSingleLine :: (Num t, Ord t) => Iv t -> Alignment -> Block a t -> SingleLine a t -> Either String (SingleLine a t)
insertSingleLine full alignment block (SL xs) = liftRight SL result where
  st@(s,t) = _block_iv block
  ivs = map _block_iv xs
  gaps = calcGaps full ivs 
  isct = filter (doesIntersect st) ivs
  b = doesFitIntoAGap st gaps
  result = if b 
    then Right (worker st block xs)
    else case alignment of
      NoAlign -> Left "does not fit into a gap"
      FitIntoGap -> case closestGap st gaps of
        Nothing -> Left "there is no intersecting gap"
        Just new -> Right $ worker new (ivmap (const new) block) xs
      ResizeLeft -> case isct of
        [ab@(a,b)] -> if a<s && b>=s && t>b
          then let new = (b,t) in Right $ worker new (ivmap (const new) block) xs
          else Left "resize/left: not good"
        _ -> Left "intersects more than one"
      ResizeRight -> case isct of
        [ab@(a,b)] -> if a<=t && b>t && s<a
          then let new = (s,a) in Right $ worker new (ivmap (const new) block) xs
          else Left "resize/right: not good"
        _ -> Left "intersects more than one"
      Move -> case isct of
        [ab@(a,b)] -> if a<s && b>=s && t>b
          then let new = (b,t+(b-s)) in if doesFitIntoAGap new gaps
            then Right $ worker new (ivmap (const new) block) xs
            else Left "does not fit"
          else if a<=t && b>t && s<a 
            then let new = (s-(t-a),a) in if doesFitIntoAGap new gaps
              then Right $ worker new (ivmap (const new) block) xs
              else Left "does not fit"
            else Left "move: not good"
        _ -> Left "intersects more than one"  
  worker uv@(u,v) block [] = [block]
  worker uv@(u,v) block bbs@(b:bs) 
    | v<=x       = block : bbs
    | otherwise  = b : worker uv block bs
    where
      xy@(x,y) = _block_iv b


insertBlock :: (Num t, Ord t) => Alignment -> Int -> Block a t -> Timeline a t -> Either String (Timeline a t)
insertBlock alignment lev block tl@(Timeline full arr) = result where
  nameArr = tlNames tl
  (mn,mx) = bounds nameArr
  forbiddenNames = concat [ nameArr ! i | i<-[mn..mx], i/=lev ]
  result = if elem (_block_name block) forbiddenNames 
    then Left "having the same name at multiple levels is forbidden"
    else liftRight (Timeline full) arr'  
  arr' = adjustElemEither (insertSingleLine full alignment block) lev arr
 
moveBlock :: (Num t, Ord t) => Alignment -> (Int, Iv t) -> (Int, Iv t) -> Maybe String -> Timeline a t -> Either String (Timeline a t)
moveBlock alignment (oldlev,oldiv) (newlev,newiv) mnewname tline = result where
  ( moldblock , temp ) = removeBlock oldlev oldiv tline  
  result = case moldblock of
    Nothing -> Left "not found"
    Just ob -> insertBlock alignment newlev newblock temp where
      newblock = case mnewname of 
        Nothing  -> ob { _block_iv = newiv }         
        Just nam -> ob { _block_iv = newiv , _block_name = nam }
        
        
-------------------------------------------------------
  
  
