
-- | A custom color selector widget for wxHaskell.

module Graphics.UI.WX.Controls.ColorWidget
  ( ColorWidget
  , VColor(..) 
  , makeColorWidget
  , getRGB
  , getRGBA
  , getAlpha
  , getColorModel
  , setRGB
  , setRGBA
  , setAlpha
  , setColorModel
  , enableColorModel
  ) where

import Control.Monad
import Control.Concurrent.MVar

import Graphics.UI.WX 
import Graphics.UI.WXCore

import Graphics.UI.WX.Controls.Knob
import Data.ColorSpaces

-------------------------------------------------------

data VColor 
  = ColRGB { unColRGB :: RGB } 
  | ColHSL { unColHSL :: HSL }
  | ColHSV { unColHSV :: HSV }

instance ColorSpace VColor where
  toRGB (ColRGB xxx) = toRGB xxx
  toRGB (ColHSL xxx) = toRGB xxx
  toRGB (ColHSV xxx) = toRGB xxx
  fromRGB = undefined
  
toTriple :: VColor -> (Float,Float,Float)
toTriple col = case col of
  ColRGB xxx -> unRGB xxx  
  ColHSL xxx -> unHSL xxx  
  ColHSV xxx -> unHSV xxx  

toList :: VColor -> [Float]  
toList col = let (a,b,c) = toTriple col in [a,b,c]

-- | ColorWidget widget. Instances: 'Widget', 'Dimensions', 'Identity', 'Tipped', 'Selection', 'Visible'. 
-- The selection attribute gives the current value.
data ColorWidget = ColorWidget 
  { _window :: StaticBox () -- Panel ()
  , _color  :: MVar VColor
  , _rgb    :: MVar RGB
  , _a      :: MVar Float
  , _knobs  :: [Knob]
  , _feedb  :: Panel ()
  , _model  :: Choice ()
  , _selectCallback :: MVar (IO ())
  }
  
-------------------------------------------------------

intToFloat :: Int -> Float
intToFloat i = fromIntegral i / 255.0

floatToInt :: Float -> Int
floatToInt x = round (x*255.0)  

-------------------------------------------------------

instance Identity ColorWidget where
  identity = mapAttrW _window identity
    
instance Widget ColorWidget where
  widget colwdg = widget (_window colwdg)    
    
instance Dimensions ColorWidget where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Tipped ColorWidget where
  tooltip = mapAttrW _window tooltip

instance Visible ColorWidget where
  visible = mapAttrW _window visible

instance Selecting ColorWidget where
  select = newEvent "select" getter setter where
    getter cw = readMVar (_selectCallback cw)
    setter cw callback = swapMVar (_selectCallback cw) callback >> return ()
    
-------------------------------------------------------

getRGB :: ColorWidget -> IO RGB
getRGB cw = readMVar (_rgb cw)

getAlpha :: ColorWidget -> IO Float
getAlpha cw = readMVar (_a cw)

getRGBA :: ColorWidget -> IO (RGB,Float)
getRGBA cw = do
  x <- readMVar (_rgb cw)
  y <- readMVar (_a   cw)
  return (x,y)
  
getColorModel :: ColorWidget -> IO Int
getColorModel cw = get (_model cw) selection

setRGB :: ColorWidget -> RGB -> IO ()
setRGB cw x = do
  swapMVar (_rgb cw) x
  changeColorModelW cw
 
setAlpha :: ColorWidget -> Float -> IO ()
setAlpha cw x = do
  swapMVar (_a cw) x 
  changeColorModelW cw

setRGBA :: ColorWidget -> RGB -> Float -> IO ()
setRGBA cw x y = do
  swapMVar (_rgb cw) x
  swapMVar (_a   cw) y
  changeColorModelW cw

setColorModel :: ColorWidget -> Int -> IO ()
setColorModel cw j = set (_model cw) [ selection := j ]

enableColorModel :: ColorWidget -> Bool -> IO ()
enableColorModel cw b = set (_model cw) [ enabled := b ]

-------------------------------------------------------
  
-- | Creates a new color widget.
makeColorWidget
  :: Window a     -- ^ parent window
  -> Bool         -- ^ alpha component
  -> [Prop ColorWidget]  -- ^ properties (eg. 'selection' or 'tooltip') 
  -> IO ColorWidget
makeColorWidget parent hasalpha props = do
  rgbmv <- newMVar (RGB (0.75,0.25,0.25))
  amv   <- newMVar 1.0
  
  callbackmv <- newMVar (return ())
  let callback = do
        tryTakeMVar callbackmv >>= \m -> case m of
          Nothing     -> return ()
          Just action -> putMVar callbackmv action >> action
  
  let alphax = if hasalpha then 50 else 0  
  win1 <- staticBoxCreate parent 0 "" (Rect 0 0 (170+alphax) 120) 0
  Size w h <- get win1 clientSize
  let (w1,h1) = (150+alphax,100)
      centerIn (w1,h1) (w,h) = Point (div (w-w1) 2) (div (h-h1) 2)
  win <- panel win1 
    [ size     := Size w1 h1 
    , position := centerIn (w1,h1) (w,h) 
    ]

  rgb <- readMVar rgbmv
  a <- readMVar amv
  let hsl = fromRGB rgb
      chsl = ColHSL hsl
  thecolor <- newMVar chsl
  
  feedback <- panel win 
    [ position := Point 5 55 
    , size     := Size (40+alphax) 40 
    , on paint := paintFeedback rgbmv amv hasalpha
    ]
    
  let mk props = makeKnob win 0 255 props :: IO Knob
      mki i = mk [ position := Point (i*50) 0 ] :: IO Knob
  knobs <- sequence [ mki i | i <- [0..3] ]
  forM_ (zip knobs (toList chsl ++ [a])) $ \(kn,x) -> set kn 
    [ selection := floatToInt x
    , on select := update rgbmv amv thecolor knobs feedback >> callback
    ]
  set (last knobs) [ visible := hasalpha ]
  
  model <- choice win  
    [ position   := Point (55+alphax) 65 
    , size       := Size 90 35
    , items      := if hasalpha then ["RGBA","HSLA","HSVA"] else ["RGB","HSL","HSV"]
    , selection  := 1
    ]
  set model [ on select := changeColorModel rgbmv amv thecolor knobs model feedback >> callback ]

  update rgbmv amv thecolor knobs feedback 
      
  let colwdg = ColorWidget win1 thecolor rgbmv amv knobs feedback model callbackmv
  set colwdg props
  return colwdg

-------------------------------------------------------

changeColorModelW w = do
  changeColorModel  
    (_rgb w)
    (_a w)
    (_color w)
    (_knobs w)
    (_model w)
    (_feedb w)
 
changeColorModel rgbmv amv thecolor knobs model feedb = do
  rgb <- readMVar rgbmv
  a <- readMVar amv
  i <- get model selection
  let new = case i of
        0 -> ColRGB (fromRGB rgb)
        1 -> ColHSL (fromRGB rgb)
        2 -> ColHSV (fromRGB rgb)
  swapMVar thecolor new      
  forM_ (zip knobs (toList new ++ [a])) $ \(knob,x) -> do
    set knob [ selection := floatToInt x ]
  repaint feedb

-- reads the knobs, and updates everything
update rgbmv amv thecolor knobs feedb = do
  oldc <- readMVar thecolor
  [a,b,c,alpha] <- liftM (map intToFloat) $ forM knobs $ \knob -> get knob selection
  let xxx = case oldc of
        ColRGB _ -> ColRGB $ RGB (a,b,c)
        ColHSL _ -> ColHSL $ HSL (a,b,c)
        ColHSV _ -> ColHSV $ HSV (a,b,c)
      rgb = toRGB xxx      
  swapMVar rgbmv rgb
  swapMVar amv alpha
  swapMVar thecolor xxx
  repaint feedb
  return ()
      
-------------------------------------------------------

paintFeedback rgbmv amv hasalpha dc rect = do
  RGB (r,g,b) <- readMVar rgbmv
  a <- readMVar amv
  let rgb = [r,g,b]
      [x ,y ,z ] = map (floatToInt                   ) rgb
      [x1,y1,z1] = map (floatToInt . \x -> x*a       ) rgb
      [x2,y2,z2] = map (floatToInt . \x -> 1+a*(x-1) ) rgb
  if not hasalpha
    then dcWithBrushStyle dc (brushSolid $ colorRGB x y z) $ dcDrawRectangle dc rect
    else do
      let Rect l t w h = rect
          rect1 = Rect l t 50 h
      dcWithPenStyle dc penTransparent $
        dcWithBrushStyle dc (brushSolid $ colorRGB x y z) $ dcDrawRectangle dc rect1
      forM_ [ (i,j) | i<-[0..3] , j<-[0..3] ] $ \(i,j) -> do
        let p = 10*i
            q = 10*j
            col = if odd (i+j) then colorRGB x1 y1 z1 else colorRGB x2 y2 z2
        dcWithPenStyle dc penTransparent $ 
          dcWithBrushStyle dc (brushSolid col) $ dcDrawRectangle dc $ 
            Rect (l+50+p) (t+q) 10 10  
      dcWithBrushStyle dc brushTransparent $ dcDrawRectangle dc rect
        
-------------------------------------------------------
  