
-- | A custom knob widget for wxHaskell.

module Graphics.UI.WX.Controls.Knob 
  ( Knob
  , makeKnob 
--  , makeKnob' 
  ) where

import Control.Monad
import Control.Concurrent.MVar

import Graphics.UI.WX 
import Graphics.UI.WXCore hiding (Event)

-------------------------------------------------------

clamp :: Int -> (Int,Int) -> Int
clamp x (minv,maxv) = max (min x maxv) minv

-- | Knob widget. Instances: 'Widget', 'Dimensions', 'Identity', 'Tipped', 'Selection', 'Visible'. 
-- The selection attribute gives the current value.
data Knob = Knob 
  { _window  :: Panel ()
  , _minmax  :: (Int,Int)
  , _value   :: MVar Int
  , _state   :: MVar KnobState
  , _handler :: MVar (IO ())
  }
  
data KnobState 
  = Up 
  | Down { _startPos :: (Int,Int) , _startValue :: Int }

-------------------------------------------------------

instance Identity Knob where
  identity = mapAttrW _window identity
    
instance Widget Knob where
  widget knob = widget (_window knob)    

instance Dimensions Knob where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Tipped Knob where
  tooltip = mapAttrW _window tooltip

instance Visible Knob where
  visible = mapAttrW _window visible
      
instance Selection Knob where
  selection = newAttr "selection" getter setter where
    getter knob   = readMVar (_value knob)
    setter knob x = swapMVar (_value knob) x >> repaint (_window knob) 

instance Selecting Knob where
  select = newEvent "select" getter setter where
    getter knob   = readMVar (_handler knob)
    setter knob x = swapMVar (_handler knob) x >> return ()

  
-------------------------------------------------------

-- default event handler
doNothing :: IO ()
doNothing = return ()

-------------------------------------------------------

-- | Creates a new knob widget.
makeKnob 
  :: Window a     -- ^ parent window
  -> Int          -- ^ minimum value
  -> Int          -- ^ maximum value; (0,127) is a sane default.
  -> [Prop Knob]  -- ^ properties (eg. 'selection' or 'tooltip') 
  -> IO Knob 
makeKnob parent minv maxv props = do
  val <- newMVar minv
  makeKnob' parent minv maxv val props 
  
-- | Creates a new knob widget.
makeKnob' 
  :: Window a     -- ^ parent window
  -> Int          -- ^ minimum value
  -> Int          -- ^ maximum value; (0,127) is a sane default.
  -> MVar Int     -- ^ a place to store the value
  -> [Prop Knob]  -- ^ properties (eg. 'selection' or 'tooltip') 
  -> IO Knob
  
makeKnob' parent minvalue maxvalue val props = do 
  win <- panelCreate parent 0 (Rect 0 0 50 50) 0 
  state <- newMVar Up

  eventhandler <- newMVar doNothing

  let minmax = (minvalue,maxvalue)
      knob = Knob win minmax val state eventhandler
      
  set win 
    [ on paint := paintKnob win minmax val  
    , on mouse := mouseKnob win knob minmax val state
    , visible  := True
    ]
    
  set knob props
  return knob

-------------------------------------------------------
  
mouseKnob :: Panel () -> Knob -> (Int,Int) -> MVar Int -> MVar KnobState -> EventMouse -> IO ()   
mouseKnob win knob minmax@(themin,themax) val state event = do
  let diff = themax-themin+1
  s <- takeMVar state
  s' <- case s of
    Up -> case event of
      MouseLeftDown (Point x y) mod -> do
        oldvalue <- readMVar val
        windowCaptureMouse win
        return $ Down (x,y) oldvalue
      _ -> return s
    Down (x0,y0) oldvalue -> case event of
      MouseLeftDrag (Point x y) mod -> do
        let sign = -1 -- if x0 < 25 then -1 else 1
            newvalue = clamp ( oldvalue + (sign * (y-y0) * diff) `div` 256 ) minmax 
        swapMVar val newvalue      
        readMVar (_handler knob) >>= id  -- call the event handler 
        return s
      MouseLeftUp (Point x y) mod -> do 
        windowReleaseMouse win
        return Up 
      _ -> return s 
  putMVar state s'         
  repaint win
    
-------------------------------------------------------
    
paintKnob :: Panel () -> (Int,Int) -> MVar Int -> DC () -> Rect -> IO ()   
paintKnob win (minvalue,maxvalue) val dc rect = do
  dcClear dc
  let center  = 25 :: Float
      radius1 = 16 :: Float
      radius2 = 21 :: Float
      n1 = 24 :: Int
      n2 = 24 :: Int
      
  let point r a = Point
        (round $ center + r * cos a) 
        (round $ center + r * sin a)

  x <- readMVar val     
  let angle = 0.75*pi + 1.5*pi* ( fromIntegral (x - minvalue) / fromIntegral (maxvalue-minvalue) )        
      txt = show x
      
  dcWithPenStyle dc (penColored (colorRGB 64 64 64) 1) $ do 
    drawLines dc
      [ point radius2 a
      | a <- [ -1.25*pi + 1.5*pi*(fromIntegral i / fromIntegral n1) | i<-[0..n1] ] 
      ] 
    drawLines dc
      [ point radius1 a
      | a <- [ -1.25*pi + 1.5*pi*(fromIntegral i / fromIntegral n2) | i<-[0..n2] ] 
      ] 
    forM_ [-1,1::Int] $ \i -> do
      let a = -0.5*pi + (fromIntegral i)*pi*0.75  
      drawLines dc
        [ point radius1 a
        , point radius2 a
        ] 

  dcWithPenStyle dc (penColored (colorRGB 160 0 0) 2) $ do 
      drawLines dc
        [ point       0 angle
        , point radius2 angle
        ] 
        
  dcWithFontStyle dc fontDefault $ do
    Size xs ys <- getTextExtent dc txt
    let pos = Point 
              (round $ center + 0    - fromIntegral xs / 2)
              (round $ center + 20   - fromIntegral ys / 2) 
    dcDrawText dc txt pos
     
-------------------------------------------------------
