
-- | Ruler widget.
-- It can be used together with the Timeline widget, for example.

{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances, FunctionalDependencies #-}
module Graphics.UI.WX.Controls.Ruler 
  ( FromDouble(..)
  , Base(..)
  , RulerWidget
  , makeRuler
  , RulerDisplay(..)
  )
  where


import Control.Monad
import Control.Concurrent.MVar

import Data.Bits
import Data.Maybe
import Data.List hiding (intersect)

import Graphics.UI.WX 
import Graphics.UI.WXCore hiding (Time)

-------------------------------------------------------

class FromDouble t where
  fromDouble :: Double -> t

instance FromDouble Int    where fromDouble = round
instance FromDouble Float  where fromDouble = realToFrac
instance FromDouble Double where fromDouble = id
 
-------------------------------------------------------

data Base = Base2 | Base10

data RulerWidget t = RulerWidget
  { _window     :: Panel ()
  , _interval   :: MVar (t,t) 
  , _base       :: MVar Base
  , _displayfun :: MVar (t -> String)    
  }
  
makeRuler :: (Real t, FromDouble t, Show t) => Window a -> Int -> (t,t) -> [Prop (RulerWidget t)] -> IO (RulerWidget t)
makeRuler parent height iv props = do
  let rect = Rect 0 0 384 height
  win <- panelCreate parent 0 rect 0

  miv   <- newMVar iv
  mbase <- newMVar Base10
  mdfun <- newMVar show

  let ruler = RulerWidget win miv mbase mdfun 
  
  set win 
    [ on paint  := paintRuler ruler
    , on resize := resizeRuler ruler
    ]

  set ruler props
  return ruler

-------------------------------------------------------

log2 :: Double -> Double
log2 x = log x / log 2

log10 :: Double -> Double
log10 x = log x / log 10

exp2 :: Double -> Double
exp2 x = exp (x * log 2)

exp10 :: Double -> Double
exp10 x = exp (x * log 10)

-------------------------------------------------------

-- Base2  -> 2 hatvanyra kerekitunk: 1,2,4,8,16,...
-- Base10 -> 1,2,5,10,20,50,100,200,500,1000 stb
roundBase :: Base -> Double -> Double
roundBase Base2 x = exp2 $ fromIntegral $ floor $ 0.5 + log2 x
roundBase Base10 x 
  | f < log10 2    =     exp10 i
  | f < log10 5    = 2 * exp10 i
  | otherwise      = 5 * exp10 i 
  where
    i = fromIntegral $ floor y
    f = y - i
    y = 0.3 + log10 x

printCentered :: DC a -> Point -> String -> IO ()
printCentered dc (Point xcenter ybottom) txt= do
  dcWithFontStyle dc fontDefault $ do
    Size xs ys <- getTextExtent dc txt
    --dcSetTextForeground dc ...
    dcDrawText dc txt $ Point (xcenter - div xs 2) (ybottom - 1 - ys)
            
paintRuler ruler@(RulerWidget win mst mbase mdfun) dc rect = do
  (s',t') <- readMVar mst
  base    <- readMVar mbase
  displayfun <- readMVar mdfun
  Size w h   <- windowGetSize win

  let 
    (s,t) = (realToFrac s', realToFrac t') :: (Double,Double)
    dur    = t - s
    fwidth = realToFrac w   :: Double
    (smallTickStride, bigmod) = case base of 
      Base2  -> let x = roundBase base $ dur / (fwidth / 20.0) in (x,4) 
      Base10 -> let x = roundBase base $ dur / (fwidth / 16.0) in (x,5)

  let background = brushSolid (colorRGB 220 200 200)
      penstyle   = penColored (colorRGB 0 0 0) 1
      
  withBrushStyle background $ \brush -> do 
    dcSetBackground dc brush
    dcClearRect dc rect

  dcWithPenStyle dc penstyle $ do
    forM_ [ ceiling ( s / smallTickStride ) .. floor ( t / smallTickStride ) ] $ \i -> do
      let time = fromIntegral i * smallTickStride 
          x = round $ (time - s) * fwidth / dur
          isbig = i `mod` bigmod == 0 
          y = if isbig then (h-16) else (h-16) `div` 2 
      dcDrawLine dc (Point x 0) (Point x y) 
      Control.Monad.when isbig $ printCentered dc (Point x h) (displayfun $ fromDouble time)
      
-------------------------------------------------------

resizeRuler ruler = do
  repaint (_window ruler)
  return ()
  
-------------------------------------------------------
  
makeAttributeFromMVar :: String -> (RulerWidget t -> MVar a) -> Attr (RulerWidget t) a
makeAttributeFromMVar name f = 
  newAttr name getter setter where
    getter ruler   = readMVar (f ruler)
    setter ruler x = do
      swapMVar (f ruler) x
      repaint (_window ruler)

class RulerDisplay ruler t | ruler -> t where 
  rulerInterval   :: Attr ruler (t,t)
  rulerBase       :: Attr ruler Base
  rulerDisplayFun :: Attr ruler (t -> String)

instance RulerDisplay (RulerWidget t) t where
  --rulerInterval :: Attr (Ruler t) (t,t)
  rulerInterval = makeAttributeFromMVar "rulerInterval" _interval 
  --rulerBase :: Attr (Ruler t) Base
  rulerBase    = makeAttributeFromMVar "rulerBase" _base
  --rulerDisplayFun :: Attr (Ruler t) (t -> String)
  rulerDisplayFun = makeAttributeFromMVar "rulerDisplayFun" _displayfun
  
-------------------------------------------------------
    
instance Widget (RulerWidget t) where
  widget ruler = widget (_window ruler)    
  
instance Dimensions (RulerWidget t) where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Sized (RulerWidget t) where
  size = mapAttrW _window size

instance Tipped (RulerWidget t) where
  tooltip = mapAttrW _window tooltip

instance Identity (RulerWidget t) where
  identity = mapAttrW _window identity  
  
-------------------------------------------------------
  