
-- | Composite widget made up from a timeline, a ruler and optionally a waveform
-- It is provided because 
--   1) the \<censored\> wxwidgets fucks up the layout all the time, really hard;
--   2) some custom resizing is needed because of the vertical scrollbar.

{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances #-}
module Graphics.UI.WX.Controls.TimelineComposite 
  ( 
    ToSamplesFun
  , TimelineComposite
  , makeTimelineComposite

  , module Graphics.UI.WX.Controls.TimelineWidget
  , module Graphics.UI.WX.Controls.Ruler
  , module Graphics.UI.WX.Controls.Waveform
  )
  where

import Control.Monad
import Control.Concurrent.MVar

import Data.Maybe
import Data.List hiding (intersect)

import Graphics.UI.WX 
import Graphics.UI.WXCore hiding (Time)

import Graphics.UI.WX.Controls.TimelineWidget
import Graphics.UI.WX.Controls.Ruler
import Graphics.UI.WX.Controls.Waveform

-------------------------------------------------------
 
-- | conversion from timeline time to samples offset in the wave
type ToSamplesFun = Int -> Int  
 
-- | the composite widget
data TimelineComposite a = TimelineComposite
  { _window  :: Panel ()
  , _tlw     :: TimelineWidget a
  , _ruler   :: RulerWidget Int
  , _wform     :: Maybe WaveformWidget
  , _tosamples :: MVar ToSamplesFun
  }
    
-------------------------------------------------------

makeTimelineComposite 
  :: Window x          -- ^ parent
  -> Int               -- ^ number of rows (levels) in the timeline (numbering starts from 1)
  -> Int               -- ^ height of a row in pixels
  -> (Int,Int)         -- ^ the time interval of the timeline
  -> Int               -- ^ height of the ruler in pixels
  -> Maybe (Int,Wave,ToSamplesFun)  -- ^ the optional waveform: (height,wave,conversion-function); 
                                    -- the function converts time to sample position in the wave 
  -> [Prop (TimelineComposite a)]                                  
  -> IO (TimelineComposite a)
makeTimelineComposite parent nrows rowheight interval rulerheight mwaveform props = do

  container <- panel parent []
  tline <- makeTimelineWidget container nrows interval rowheight [ zoom := 1 ] -- :: IO (TimelineWidget a)  
  ruler <- makeRuler container rulerheight interval [] :: IO (RulerWidget Int)
  (mwform,mtosamples) <- case mwaveform of
    Just (wfh,wave,tosamples) -> do
      mwf <- liftM Just $ makeWaveformWidget container wfh wave []
      mts <- newMVar tosamples   
      return (mwf,mts)
    Nothing -> newEmptyMVar >>= \mv -> return (Nothing,mv)
    
  let composite = TimelineComposite container tline ruler mwform mtosamples
    
  set container 
    [ on resize := do
        Size w h <- get container size
        set ruler [ position := Point 0 0 ] 
        case mwaveform of
          Just (wformheight,_,_) -> do
            set tline [ position := Point 0 rulerheight , size := Size w (h-rulerheight-wformheight) ]
            set (fromJust mwform) [ position := Point 0 (h-wformheight) ]
          Nothing -> do
            set tline [ position := Point 0 rulerheight , size := Size w (h-rulerheight) ]
    ]
  
  set tline 
    [ onZoomScroll := \z st@(s,t) (Size w _) -> do
        --logStatus ("zoomScroll: " ++ show z ++ " " ++ show st)
        set ruler [ size := Size w rulerheight , rulerInterval := st ]
        case mwform of
          Nothing -> return ()
          Just wform -> do
            toSamples <- readMVar mtosamples
            let (wformheight,_,_) = fromJust mwaveform
            set wform [ size := Size w wformheight , waveformInterval := (toSamples s, toSamples t) ]
    ]
    
  set composite props  
  return composite       

-------------------------------------------------------

instance TimelineDisplay (TimelineComposite a) a where
  feedback       = mapAttrW _tlw feedback
  onBlockSelect  = mapAttrW _tlw onBlockSelect 
  onBlockMove    = mapAttrW _tlw onBlockMove
  onBlockResize  = mapAttrW _tlw onBlockResize
  onBlockInsert  = mapAttrW _tlw onBlockInsert
  onBlockInsertSuccess = mapAttrW _tlw onBlockInsertSuccess
  onZoomScroll   = mapAttrW _tlw onZoomScroll
  nrows    = mapAttrW _tlw nrows
  duration = mapAttrW _tlw duration
  blocks   = mapAttrW _tlw blocks
  zoom     = mapAttrW _tlw zoom

toSamplesFun :: Attr (TimelineComposite a) ToSamplesFun
toSamplesFun = newAttr "toSamplesFun" getter setter where
  getter w   = readMVar (_tosamples w)
  setter w x = do 
    swapMVar (_tosamples w) x 
    case (_wform w) of
      Just wf -> repaint wf
      Nothing -> return ()
      
instance RulerDisplay (TimelineComposite a) Int where
  rulerInterval   = mapAttrW _ruler rulerInterval
  rulerBase       = mapAttrW _ruler rulerBase 
  rulerDisplayFun = mapAttrW _ruler rulerDisplayFun

maybeMapAttrW s f attr = newAttr s getter setter where
  getter x = case f x of 
    Just w  -> get w attr
    Nothing -> error "maybeMapAttrW/getter"
  setter x y = case f x of
    Just w  -> set w [ attr := y ]
    Nothing -> error "maybeMapAttrW/setter"
  
instance WaveDisplay (TimelineComposite a) where
  waveformInterval = maybeMapAttrW "waveformInterval" _wform waveformInterval
  waveformWave     = maybeMapAttrW "waveformWave"     _wform waveformWave
  
-------------------------------------------------------
       
instance Widget (TimelineComposite a) where
  widget ruler = widget (_window ruler)    
  
instance Dimensions (TimelineComposite a) where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Sized (TimelineComposite a) where
  size = mapAttrW _window size

instance Tipped (TimelineComposite a) where
  tooltip = mapAttrW _window tooltip

instance Identity (TimelineComposite a) where
  identity = mapAttrW _window identity  
  
-------------------------------------------------------

