
-- | Waveform widget.
-- It can be used together with the Timeline widget, for example.

{-# LANGUAGE MultiParamTypeClasses #-}
module Graphics.UI.WX.Controls.Waveform 
  ( Wave(..)
  , WaveformWidget
  , makeTestWave
  , makeWaveformWidget
  , WaveDisplay(..)
  )
  where

import Control.Monad
import Control.Concurrent.MVar

import Data.Bits
import Data.Word
import Data.Maybe
import Data.List hiding (intersect)

import Graphics.UI.WX 
import Graphics.UI.WXCore hiding (Time)
 
import Foreign
import Foreign.ForeignPtr 
 
-------------------------------------------------------

data Wave = Wave
  { _nsamples  :: Int
  , _isStereo  :: Bool
  , _rawdata   :: ForeignPtr Float
  }
    
data WaveformWidget = WaveformWidget
  { _window     :: Panel ()
  , _interval   :: MVar (Int,Int)    -- in samples 
  , _wave       :: MVar Wave
  }
  
-- | Creates a test wave. Then argument gives the length in secs.
makeTestWave :: Int -> IO Wave
makeTestWave secs = do
  let len = secs * 44100
      f :: Float -> Float
      f x = 0.2 * sin (x / 400 + 0.001)
          + 0.3 * cos (x / 402 + 0.104)
          - 0.4 * sin (x / 797 + 0.467)
          - 0.3 * sin (x / 40000)
  stuff <- mallocForeignPtrArray len :: IO (ForeignPtr Float) 
  withForeignPtr stuff $ \p ->
    forM_ [0..len-1] $ \i -> pokeElemOff p i (f $ fromIntegral i {-/ 44100.0-} ) 
  return $ Wave len False stuff
  
makeWaveformWidget :: Window a -> Int -> Wave -> [Prop WaveformWidget] -> IO WaveformWidget
makeWaveformWidget parent height wave props = do
  let rect = Rect 0 0 384 height
  win <- panelCreate parent 0 rect 0

  miv  <- newMVar (0,_nsamples wave)
  mwav <- newMVar wave 

  let wform = WaveformWidget win miv mwav 
  
  set win 
    [ on paint  := paintWaveformWidget wform
    , on resize := resizeWaveformWidget wform
    ]

  set wform props
  return wform

-------------------------------------------------------
            
clamp :: Float -> Float
clamp x = min 1.0 (max (-1.0) x)

{-# INLINE pixel #-}
pixel :: Ptr Word8 -> Int -> Int -> Int -> Int -> IO ()            
pixel p w h xc yc = Control.Monad.when (xc>=0 && xc<w && yc>=0 && yc<h) $ do
  let q = p `plusPtr` ((xc+yc*w)*3)
  poke  q              (0  ::Word8)
  poke (q `plusPtr` 1) (0  ::Word8)
  poke (q `plusPtr` 2) (100::Word8)
      
paintWaveformWidget wform@(WaveformWidget win mst mwave) dc rect = do
  (s,t) <- readMVar mst
  wave  <- readMVar mwave
  Size w h <- windowGetSize win

  let r = 0.5 * fromIntegral h :: Float
      hh = floor r 
      dur = fromIntegral $ t - s :: Double
      durperw = dur / fromIntegral w
      f :: Int -> Int
      f i = s + ( round $ fromIntegral i * durperw )
      n = _nsamples wave
      g = case (_isStereo wave) of
        True -> \p i -> do 
          let j = f i 
              k = j+j
          if j<n && j>=0
            then do 
              a <- peekElemOff p (k  )
              b <- peekElemOff p (k+1)
              return $ round $  r + r * clamp (0.5*(a+b)) 
            else 
              return hh  
        False -> \p i -> do
          let j = f i
          if j<n && j>=0
            then do 
              let j = f i
              a <- peekElemOff p j
              return $ round $ r + r * clamp a 
            else
              return hh

  withForeignPtr (_rawdata wave) $ \p ->  do
    allocaBytes (3*w*h) $ \q -> do 

      forM_ [ 0..w*h-1 ] $ \i -> do
        let j = i+i+i
        poke (q `plusPtr`  j   ) (160::Word8)
        poke (q `plusPtr` (j+1)) (180::Word8)
        poke (q `plusPtr` (j+2)) (200::Word8)

      forM_ [ 0..w-1 ] $ \i -> do
        y <- g p i
        if y<hh 
          then forM_ [y..hh] (pixel q w h i)
          else forM_ [hh..y] (pixel q w h i)
    
      image <- imageCreateFromData (Size w h) q
      bitmap <- bitmapCreateFromImage image 8
      dcDrawBitmap dc bitmap (Point 0 0) False
    
{-              
  let background = brushSolid (colorRGB 180 200 220)
      penstyle   = penColored (colorRGB 00 0 100) 1
      
  withBrushStyle background $ \brush -> do 
    dcSetBackground dc brush
    dcClearRect dc rect

  withForeignPtr (_rawdata wave) $ \p ->  
    dcWithPenStyle dc penstyle $ do
      forM_ [ 0..w-1 ] $ \i -> do
        y <- g p i
        
        -- OH. MY. FUCKING. GOD.
        -- drawing a PIXEL seems to be a way too hard task... (OSX) 
        -- dcDrawPoint dc (Point i y) 
        dcDrawLine dc (Point i hh) (Point i y) 
-}
      
-------------------------------------------------------

resizeWaveformWidget wform = do
  repaint (_window wform)
  return ()
  
-------------------------------------------------------
  
makeAttributeFromMVar :: String -> (WaveformWidget -> MVar a) -> Attr WaveformWidget a
makeAttributeFromMVar name f = 
  newAttr name getter setter where
    getter wform   = readMVar (f wform)
    setter wform x = do
      swapMVar (f wform) x
      repaint (_window wform)

class WaveDisplay widg where
  waveformInterval :: Attr widg (Int,Int)
  waveformWave     :: Attr widg Wave

instance WaveDisplay WaveformWidget where
  --waveformInterval :: Attr WaveformWidget (Int,Int)
  waveformInterval = makeAttributeFromMVar "waveformInterval" _interval 
  --waveformWave :: Attr WaveformWidget Wave
  waveformWave = makeAttributeFromMVar "waveformWave" _wave

-------------------------------------------------------
    
instance Widget WaveformWidget where
  widget wform = widget (_window wform)    

instance Dimensions WaveformWidget where
  outerSize   = mapAttrW _window outerSize
  position    = mapAttrW _window position
  area        = mapAttrW _window area
  bestSize    = mapAttrW _window bestSize  
  clientSize  = mapAttrW _window clientSize
  virtualSize = mapAttrW _window virtualSize

instance Paint WaveformWidget where
  paint     = error "paint/WaveformWidget"
  paintRaw  = error "paintRaw/WaveformWidget"
  repaint w = repaint (_window w)

instance Sized WaveformWidget where
  size = mapAttrW _window size

instance Tipped WaveformWidget where
  tooltip = mapAttrW _window tooltip

instance Identity WaveformWidget where
  identity = mapAttrW _window identity  
  
-------------------------------------------------------
  